# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union
from collections.abc import Iterable

from abc import ABCMeta

import jax.numpy as jnp
from jax.stages import ArgInfo

from netket.hilbert import AbstractHilbert
from netket.utils.types import Array
from netket.jax import canonicalize_dtypes
from netket.utils.numbers import is_scalar

from netket.operator._abstract_operator import AbstractOperator
from netket.operator._discrete_operator import DiscreteOperator
from netket.operator._discrete_operator_jax import DiscreteJaxOperator
from netket.operator._continuous_operator import ContinuousOperator


def _flatten_sumoperators(operators: Iterable[AbstractOperator], coefficients: Array):
    """Flatten sumoperators inside of operators."""
    new_operators = []
    new_coeffs = []
    for op, c in zip(operators, coefficients):
        if isinstance(op, SumOperator):
            new_operators.extend(op.operators)
            new_coeffs.extend(c * op.coefficients)
        else:
            new_operators.append(op)
            new_coeffs.append(c)
    return new_operators, new_coeffs


class SumOperatorMeta(ABCMeta):
    def __call__(cls, *args, **kwargs):
        __tracebackhide__ = True
        # This logic overrides the constructor, such that if someone tries to
        # construct this class directly by calling `SumOperator(...)`
        # it will construct either a DiscreteHilbert or TensorDiscreteHilbert
        from .operator import SumGenericOperator
        from .discrete_operator import SumDiscreteOperator
        from .discrete_jax_operator import SumDiscreteJaxOperator
        from .continuous import SumContinuousOperator

        if cls is SumOperator:
            if all(isinstance(op, DiscreteJaxOperator) for op in args):
                cls = SumDiscreteJaxOperator
            elif all(isinstance(op, DiscreteOperator) for op in args):
                cls = SumDiscreteOperator
            elif all(isinstance(op, ContinuousOperator) for op in args):
                cls = SumContinuousOperator
            else:
                cls = SumGenericOperator
            return cls(*args, **kwargs)
        else:
            # If this is the Metaclass of a subclass, dispatch directly
            return super().__call__(*args, **kwargs)


class SumOperator(metaclass=SumOperatorMeta):
    """
    Base class for sum of quantum operators.

    This class represents a linear combination of quantum operators with coefficients,
    implementing the mathematical concept of :math:`\\sum_i c_i \\hat{H}_i` where
    :math:`c_i` are scalar coefficients and :math:`\\hat{H}_i` are quantum operators.

    The class uses a metaclass dispatch mechanism to automatically select the
    appropriate specialized subclass based on the types of operators being summed:

    * :class:`~netket.operator._sum.discrete_jax_operator.SumDiscreteJaxOperator`
      for sums of :class:`~netket.operator.DiscreteJaxOperator` instances
    * :class:`~netket.operator._sum.discrete_operator.SumDiscreteOperator`
      for sums of :class:`~netket.operator.DiscreteOperator` instances
    * :class:`~netket.operator._sum.continuous.SumContinuousOperator`
      for sums of :class:`~netket.operator.ContinuousOperator` instances
    * :class:`~netket.operator._sum.operator.SumGenericOperator`
      for mixed operator types or fallback cases.
    """

    _operators: tuple[ContinuousOperator, ...]
    _coefficients: Array
    _is_hermitian: bool

    def __init__(
        self,
        operators: Iterable[AbstractHilbert],
        *args,
        coefficients: Union[float, Iterable[float]] = 1.0,
        dtype=None,
        **kwargs,
    ):
        r"""Constructs a Sum of Operators.

        Args:
            *operators: An iterable of quantum operators to be summed. All operators
                must act on the same Hilbert space.
            coefficients: Scalar coefficient or iterable of coefficients for each
                operator. If a single scalar is provided, it will be used for all
                operators. Default is 1.0.
            dtype: Data type for the coefficients. If None, it will be inferred
                from the operators and coefficients.
        """
        hi_spaces = [op.hilbert for op in operators]
        if not all(hi == hi_spaces[0] for hi in hi_spaces):
            raise NotImplementedError(
                "Cannot construct a SumOperator for operators on different Hilbert Spaces"
            )

        if is_scalar(coefficients):
            coefficients = [coefficients for _ in operators]

        # ArgInfo shows up in packing with .lower() sometimes... it breaks all
        if not isinstance(coefficients, ArgInfo):
            if len(operators) != len(coefficients):
                raise AssertionError("Each operator needs a coefficient")

            operators, coefficients = _flatten_sumoperators(operators, coefficients)

            dtype = canonicalize_dtypes(*operators, *coefficients, dtype=dtype)
            coefficients = jnp.asarray(coefficients, dtype=dtype)

        self._operators = tuple(operators)
        self._coefficients = coefficients

        # Call parent classes without dtype parameter - we handle dtype ourselves
        super().__init__(
            *args, **kwargs
        )  # forwards all unused arguments so that this class is a mixin.

        # Set our computed dtype after parent __init__ methods have run
        # This ensures our dtype takes precedence over any parent class dtype computation
        self._dtype = dtype

    @property
    def dtype(self):
        """The data type of the operator coefficients.

        Returns:
            The numpy/JAX dtype used for the coefficients array.
        """
        return self._dtype

    @property
    def operators(self) -> tuple[AbstractOperator, ...]:
        r"""The tuple of all operators in the terms of this sum.

        Each operator corresponds to a term :math:`\hat{O}_i` in the sum
        :math:`\sum_i c_i \hat{O}_i`, where each operator is multiplied by
        its corresponding coefficient from :attr:`coefficients`.

        Returns:
            A tuple containing all the quantum operators being summed.
        """
        return self._operators

    @property
    def coefficients(self) -> Array:
        r"""The coefficients for each operator term in the sum.

        Each coefficient corresponds to :math:`c_i` in the sum
        :math:`\sum_i c_i \hat{O}_i`, where each coefficient multiplies
        its corresponding operator from :attr:`operators`.

        Returns:
            A JAX array containing the scalar coefficients for each operator.
        """
        return self._coefficients

    def __repr__(self) -> str:
        strs = [f"{type(self).__name__} with terms:"]
        for op, c in zip(self.operators, self.coefficients):
            op_str = str(op).splitlines()
            # Format the first line with the coefficient and operator
            formatted_op = f" ∙ {c} * {op_str[0]}"
            # Add subsequent lines with proper indentation
            if len(op_str) > 1:
                formatted_op += "\n" + "\n".join(
                    f"   {' ' * len(str(c))}   {line}" for line in op_str[1:]
                )
            strs.append(formatted_op)
        return "\n".join(strs)

    def __add__(self, other):
        if not isinstance(other, AbstractOperator):
            return NotImplemented

        if isinstance(other, SumOperator):
            ops = self.operators + other.operators
            coeffs = jnp.concatenate([self.coefficients, other.coefficients])
            dtype = self.dtype if self.dtype == other.dtype else None
        else:
            ops = (*self.operators, other)
            coeffs = jnp.concatenate([self.coefficients, jnp.array([1.0])])
            dtype = self.dtype if self.dtype == other.dtype else None

        return SumOperator(*ops, coefficients=coeffs, dtype=dtype)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from collections.abc import Hashable, Iterable


from netket.utils import HashableArray, struct
from netket.utils.types import DType, PyTree, Array
from netket.operator import ContinuousOperator

from netket.operator._sum.base import SumOperator, SumOperatorMeta


class SumContinuousMeta(struct.pytree.PytreeMeta, SumOperatorMeta):
    """
    Workaround to allow SumContinuousOperator to be a subclass of both
    ContinuousOperator and SumOperator, because the first is a
    Pytree and the latter has the custom dispatch logic.
    """

    pass


class SumContinuousOperator(
    SumOperator, ContinuousOperator, metaclass=SumContinuousMeta
):
    r"""This class implements the action of the _expect_kernel()-method of
    ContinuousOperator for a sum of ContinuousOperator objects.
    """

    _operators: tuple[ContinuousOperator, ...]
    _coefficients: Array
    _is_hermitian: bool = struct.static_field()

    def __init__(
        self,
        *operators: tuple[ContinuousOperator, ...],
        coefficients: float | Iterable[float] = 1.0,
        dtype: DType | None = None,
    ):
        r"""
        Returns the action of a sum of local operators.
        Args:
            operators: A list of ContinuousOperator objects
            coefficients: A coefficient for each ContinuousOperator object
            dtype: Data type of the coefficients
        """
        self._is_hermitian = all([op.is_hermitian for op in operators])
        super().__init__(
            operators, operators[0].hilbert, coefficients=coefficients, dtype=dtype
        )

    @property
    def is_hermitian(self) -> bool:
        return self._is_hermitian

    def _expect_kernel(
        self,
        logpsi: Callable,
        params: PyTree,
        x: Array,
    ) -> Array:
        result = [
            self.coefficients[i] * op._expect_kernel(logpsi, params, x)
            for i, op in enumerate(self.operators)
        ]
        return sum(result)

    @struct.property_cached(pytree_ignore=True)
    def _attrs(self) -> tuple[Hashable, ...]:
        return (  # type: ignore
            self.hilbert,
            self.operators,
            HashableArray(self.coefficients),
            self.dtype,
        )



from functools import partial

import jax
import jax.numpy as jnp
from jax.util import safe_zip, safe_map


def _tree_map_multi_out(f, tree, *rest, is_leaf=None):
    leaves, treedef = jax.tree_util.tree_flatten(tree, is_leaf)
    all_leaves = [leaves] + [treedef.flatten_up_to(r) for r in rest]
    res = [f(*xs) for xs in safe_zip(*all_leaves)]
    return tuple(safe_map(treedef.unflatten, safe_zip(*res)))


def _treeify(f):
    def _f(x, *args, **kwargs):
        return _tree_map_multi_out(lambda y: f(y, *args, **kwargs), x)

    return _f


def _unchunk(x, _x_rest=None):
    res = jax.tree_util.tree_map(
        partial(jax.lax.collapse, start_dimension=0, stop_dimension=2), x
    )
    if _x_rest is not None:
        return jax.tree_util.tree_map(
            lambda *a: jnp.concatenate(a, axis=0), res, _x_rest
        )
    else:
        return res


@_treeify
def _chunk(x, chunk_size=None):
    # chunk_size=None -> add just a dummy chunk dimension, same as np.expand_dims(x, 0)
    if x.ndim == 0:
        raise ValueError("x cannot be chunked as it has 0 dimensions.")
    n = x.shape[0]
    if chunk_size is None:
        chunk_size = n

    n_chunks, n_rest = divmod(n, chunk_size)
    x_chunks = x[: n - n_rest, ...].reshape((n_chunks, chunk_size) + x.shape[1:])
    x_rest = x[n - n_rest :, ...]
    return x_chunks, x_rest


def _chunk_size(x):
    b = set(map(lambda x: x.shape[:2], jax.tree_util.tree_leaves(x)))
    if len(b) != 1:
        raise ValueError(
            "The arrays in x have inconsistent chunk_size or number of chunks"
        )
    return b.pop()[1]


def _chunk_without_rest(x, chunk_size):
    x_chunks, x_rest = _chunk(x, chunk_size)

    # error if we have rest
    def _check(x, x_rest):
        if x_rest.size != 0:
            raise ValueError(
                "The first dimension of x must be divisible by chunk_size."
                + f"\nGot x.shape={x.shape} but chunk_size={chunk_size}."
            )
        return x

    jax.tree.map(_check, x, x_rest)
    return x_chunks


def unchunk(x_chunked):
    """
    Merge the first two axes of an array (or a pytree of arrays)
    Args:
        x_chunked: an array (or pytree of arrays) of at least 2 dimensions
    Returns: a pair (x, chunk_fn)
        where x is x_chunked reshaped to (-1,)+x.shape[2:]
        and chunk_fn is a function which restores x given x_chunked
    """
    return _unchunk(x_chunked), partial(
        _chunk_without_rest, chunk_size=_chunk_size(x_chunked)
    )


def chunk(x, chunk_size=None):
    """
    Split an array (or a pytree of arrays) into chunks along the first axis

    Args:
        x: an array (or pytree of arrays)
        chunk_size: an integer or None (default)
            The first axis in x must be a multiple of chunk_size
    Returns: a pair (x_chunked, unchunk_fn) where
        - x_chunked is x reshaped to (-1, chunk_size)+x.shape[1:]
          if chunk_size is None then it defaults to x.shape[0], i.e. just one chunk
        - unchunk_fn is a function which restores x given x_chunked
    """

    return _chunk_without_rest(x, chunk_size), _unchunk



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""General-purpose internal utilities.

If a function or class is used in multiple modules, put it here.
"""

import functools
import inspect
import operator
from typing import Any, TypeVar, Union
from collections.abc import Callable, Iterable, Sequence, Sized
import warnings

import jax
from jax import core
from jax import random
import jax.numpy as jnp
from jax.tree_util import tree_all
from jax.tree_util import tree_map
import numpy as np


PyTree = Any


Axes = Union[int, Sequence[int]]


def is_list_or_tuple(x) -> bool:
    # We do not want to return True if x is a subclass of list or tuple since
    # otherwise this will return true for namedtuples.
    return type(x) == list or type(x) == tuple


def nt_tree_fn(
    nargs: int | None = None,
    tree_structure_argnum: int | None = None,
    reduce: Callable = lambda x: x,
):
    """Convert a function that acts on single inputs to one that acts on trees.

    `nt_tree_fn` treats the first `nargs` arguments as NTTrees and the remaining
    arguments as broadcasted over the tree structure. `nt_tree_fn` then calls the
    function on each leaf of the tree. Each node of the tree optionally calls a
    reduce function over the values of its children.

    If `tree_structure_argnum` is None then each of the NTTrees must have the same
    structure. If `tree_structure_argnum` is an integer then a specific tree is
    used to infer the structure.

    Args:
      nargs:
        The number of arguments to be treated as NTTrees. If `nargs` is `None`
        then all the arguments are used. `nargs` can also be negative which
        follows numpy's semantics for array indexing.

      tree_structure_argnum:
        The argument used to infer the tree structure to be traversed. If
        `tree_structure_argnum` is None then a check is performed to ensure that
        all trees have the same structure.

      reduce:
        A callable that is applied recursively by each internal tree node to its
        children.

    Returns:
      A decorator `tree_fn` that transforms a function, `fn`, from acting on
      leaves to acting on NTTrees.
    """

    def check_tree_structure(args):
        """Ensure the structure of the trees in each of the `nargs` is the same."""
        if any(is_list_or_tuple(x) for x in args):
            if not all(type(x) == type(args[0]) for x in args[1:]):
                raise TypeError(
                    f"Inconsistent NTTree structure found. "
                    f"Node Types: {[type(x) for x in args]}."
                )
            for x in zip(*args):
                # Regarding the use of zip, consider an example `x1 = x2 = (1, (1, 1))`.
                # We would like to determine whether these two trees have the same
                # structure.

                # On the first recurrence `x1` and `x2` are both tuples so the check
                # passes and `zip(*args) = [(1, 1), ((1, 1), (1, 1))]` so that
                # `(check_tree_structure(x) for x in zip(x1, x2))` will first check that
                # the first element of `x1` has the same tree structure as the first
                # element of `x2` and then the second element and so on.
                check_tree_structure(x)

    def tree_fn(fn):
        @wraps(fn)
        def wrapped_fn(*args, **kwargs):
            _nargs = len(args) if nargs is None else nargs
            recurse, norecurse = args[:_nargs], args[_nargs:]

            structure_argnum = tree_structure_argnum
            if structure_argnum is None:
                check_tree_structure(recurse)
                structure_argnum = 0

            if is_list_or_tuple(args[structure_argnum]):
                list_or_tuple = type(args[structure_argnum])
                return reduce(
                    list_or_tuple(
                        wrapped_fn(*(xs + norecurse), **kwargs) for xs in zip(*recurse)
                    )
                )
            return fn(*args, **kwargs)

        return wrapped_fn

    return tree_fn


def all_none(x, attr: str | None = None) -> bool:
    get_fn = (lambda x: x) if attr is None else lambda x: getattr(x, attr)
    return tree_all(tree_map(lambda x: get_fn(x) is None, x))


def wraps(f):
    def wrapper(g):
        @functools.wraps(f)
        def h(*args, **kwargs):
            return g(*args, **kwargs)

        h.__signature__ = inspect.signature(f)
        return h

    return wrapper


@nt_tree_fn(nargs=2, reduce=lambda x: jnp.all(jnp.array(x)))
def x1_is_x2(
    x1: jnp.ndarray, x2: jnp.ndarray | None = None, eps: float = 1e-12
) -> bool | jnp.ndarray:
    if not isinstance(x1, (np.ndarray, jnp.ndarray)):
        raise TypeError(f"`x1` must be an ndarray. A {type(x1)} is found.")

    if x2 is None:
        return True

    if x1 is x2:
        return True

    if x1.shape != x2.shape:
        return False

    if jax.default_backend() == "tpu":
        eps = 1e-4

    try:
        diff = x1 - x2
    except TypeError:
        # inputs are e.g. custom PRNGKeys which don't define subtraction.
        return jnp.all(x1 == x2)
    else:
        return jnp.all(jnp.abs(diff) < eps)


def _get_ndim(x: int | Sized | jnp.ndarray) -> int:
    """Get number of dimensions given number of dimensions / shape / array."""
    if hasattr(x, "ndim"):
        n = x.ndim
    elif hasattr(x, "__len__"):
        n = len(x)
    elif isinstance(x, int):
        n = x
    else:
        raise TypeError(x, type(x))
    return n


def canonicalize_axis(axis: Axes, x: int | Sized | jnp.ndarray) -> list[int]:
    """Converts axis into a sorted non-negative list.

    Args:
      axis: input axis.
      x: array / shape / number of dimensions.

    Returns:
      A sorted list of integer axes.
    """
    axis = [axis] if isinstance(axis, int) else list(axis)
    n = _get_ndim(x)
    return list(set(np.arange(n)[axis]))


def zip_axes(
    x: jnp.ndarray, start_axis: int = 0, end_axis: int | None = None
) -> jnp.ndarray:
    """Zip (interleave) axes starting from `start_axis`.

    Changes the shape as follows:
    `[..., X, Y, Z, ..., X, Y, Z, ...] -> [..., X, X, ..., Y, Y, ..., Z, Z, ...]`

    Args:
      x: `jnp.ndarray` with an even number of dimensions following `start_axis`.
      start_axis: `int`, number of axis from which to zip (interleave).
      end_axis: `int`, number of axis until which to zip (interleave).

    Returns:
      A `jnp.ndarray` with a new shape.
    """
    return _zip_axes(x, start_axis, end_axis, unzip=False)


def unzip_axes(
    x: jnp.ndarray, start_axis: int = 0, end_axis: int | None = None
) -> jnp.ndarray:
    """Unzip (de-interleave) axes starting from `start_axis`.

    Changes the shape as follows:
    `[..., X, X, ..., Y, Y, ..., Z, Z, ...] -> [..., X, Y, Z, ..., X, Y, Z, ...]`

    Args:
      x: `jnp.ndarray` with an even number of dimensions following `start_axis`.
      start_axis: `int`, number of axis from which to unzip (de-interleave).
      end_axis: `int`, number of axis until which to unzip (de-interleave).

    Returns:
      A `jnp.ndarray` with a new shape.
    """
    return _zip_axes(x, start_axis, end_axis, unzip=True)


def _zip_axes(
    x: jnp.ndarray,
    start_axis: int = 0,
    end_axis: int | None = None,
    unzip: bool = False,
) -> jnp.ndarray:
    """Zip/unzip (interleave/de-interleave) axes starting from `start_axis`.

    Changes the shape as follows:
      If `unzip == True`:
      `[..., X, X, ..., Y, Y, ..., Z, Z, ...] -> [..., X, Y, Z, ..., X, Y, Z, ..]`
      If `unzip == False`:
      `[..., X, Y, Z, ..., X, Y, Z, ...] -> [..., X, X, ..., Y, Y, ..., Z, Z, ..]`

    Args:
      x: `jnp.ndarray` with an even number of dimensions following `start_axis`.
      start_axis: `int`, number of axis from which to zip/unzip.
      end_axis: `int`, number of axis until which to zip/unzip.
      unzip: `bool`, set to `True` to unzip instead of zip.

    Returns:
      A `jnp.ndarray` with a new shape.
    """
    if end_axis is None:
        end_axis = x.ndim

    half_ndim, ragged = divmod(end_axis - start_axis, 2)
    if ragged:
        raise ValueError(
            f"Need even number of axes to zip, got {end_axis - start_axis}."
        )

    odd_axes = range(start_axis + 1, end_axis, 2)
    last_axes = range(end_axis - half_ndim, end_axis)

    if unzip:
        x = jnp.moveaxis(x, odd_axes, last_axes)
    else:
        x = jnp.moveaxis(x, last_axes, odd_axes)
    return x


_ArrayOrShape = TypeVar(
    "_ArrayOrShape",
    np.ndarray,
    jnp.ndarray,
    list[int],
    tuple[int, ...],
)


def size_at(
    x: _ArrayOrShape | core.ShapedArray, axes: Iterable[int] | None = None
) -> int:
    if hasattr(x, "shape"):
        x = x.shape

    if axes is None:
        axes = range(len(x))

    return functools.reduce(operator.mul, [x[a] for a in axes], 1)


def _read_keys(key, x1, x2):
    """Read dropout key.

    `key` might be a tuple of two rng keys or a single rng key or None. In
    either case, `key` will be mapped into two rng keys `key1` and `key2` to
    make sure `(x1==x2) == (key1==key2)`.
    """

    if key is None or all_none(x2):
        key1 = key2 = key
    elif isinstance(key, tuple) and len(key) == 2:
        key1, key2 = key
        new_key = jnp.where(x1_is_x2(key1, key2), random.fold_in(key2, 1), key2)
        key2 = jnp.where(x1_is_x2(x1, x2), key1, new_key)
        warnings.warn(
            "The value of `key[1]` might be replaced by a new value if "
            "key[0] == key[1] and x1 != x2 or key[0] != key[1] and "
            "x1 == x2."
        )
    elif isinstance(key, jnp.ndarray):
        key1 = key
        key2 = jnp.where(x1_is_x2(x1, x2), key1, random.fold_in(key, 1))
    else:
        raise TypeError(type(key))
    return key1, key2


def split_kwargs(kwargs, x1=None, x2=None):
    """Splitting `kwargs`.

    Specifically,
      1. if kwarg is a rng key, it will be split into two keys.
      2. else if it is a tuple of length two, the tuple will be split into two
         parts, one for kwargs1 and the other for kwargs2.
      3. else it is copied to kwargs1 and kwargs2.

    """
    kwargs1 = {}
    kwargs2 = {}
    for k, v in kwargs.items():
        if x2 is not None and k == "rng":
            key1, key2 = _read_keys(v, x1, x2)
            kwargs1[k] = key1
            kwargs2[k] = key2
        elif isinstance(v, tuple) and len(v) == 2:
            kwargs1[k] = v[0]
            kwargs2[k] = v[1]
        else:
            kwargs1[k] = kwargs2[k] = v

    return kwargs1, kwargs2



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, TYPE_CHECKING
from collections.abc import Callable


import orjson

import numpy as np

import jax

from typing import Self

if TYPE_CHECKING:
    from .history import History
else:
    History = Any


def _format_key(k):
    if hasattr(k, "key"):
        return str(k.key)
    elif hasattr(k, "idx"):
        return str(k.idx)
    elif hasattr(k, "name"):
        return str(k, k.name)
    else:
        return str(k)  # fallback


class HistoryDict:
    """A class that behaves like a dictionary, mapping strings to History instances or nested
    HistoryDict instances.

    This class behaves to all effects as a dictionary at runtime, but it allows us to specify
    custom serialization and deserialization routines, which are used to save and load the data.

    .. note::

        This class can be used to deserialize a dictionary of histories, such as the standard
        log files obtained from :class:`~netket.logging.RuntimeLog` and :class:`~netket.logging.JsonLog`.


    .. warning::

        This class should not be constructed directly. Instead, use
        :func:`~netket.utils.history.accum_histories_in_tree` to build complex nested
        history structures, or create instances through :class:`~netket.logging.RuntimeLog`
        for logging purposes.

    """

    def __init__(self, *args, **kwargs):
        """
        Create a new HistoryDict instance from a dictionary of histories.
        """
        data = dict(*args, **kwargs)
        # Do not nest HistoryDict inside Historydict!
        for k, d in data.items():
            if isinstance(d, HistoryDict):
                data[k] = d.to_dict()
        self._data = data

    def __setitem__(self, key: str, value: History | Self):
        if isinstance(value, HistoryDict):
            value = value.to_dict()
        self._data[key] = value

    def __getitem__(self, key: str, *, wrap_dicts=True) -> History | Self:
        if key not in self._data and "/" in key:
            keys = key.split("/")
            val = self._data
            for subkey in keys:
                val = val[subkey]  # Traverse through subdictionaries
            if wrap_dicts and isinstance(val, dict):
                return HistoryDict(val)
            return val
        val = self._data[key]
        if wrap_dicts and isinstance(val, dict):
            return HistoryDict(val)
        return val

    def __repr__(self):
        dk, _ = jax.tree_util.tree_flatten_with_path(self._data)
        if len(dk) == 0:
            return "HistoryDict({})"

        _repr_str = f"HistoryDict with {len(dk)} elements:"
        for keys, val in dk:
            path = "/".join([_format_key(k) for k in keys])
            _repr_str += f"\n\t'{path}' -> " + str(val)

        return _repr_str

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def keys(self):
        return self._data.keys()

    def items(self):
        return self._data.items()

    def values(self):
        return self._data.values()

    def get(self, key: str, default=None):
        """
        Get a value from the dictionary, with a default value if the key is not present.

        If the value is a dictionary, it is wrapped in a HistoryDict instance.

        Args:
            key: The key to retrieve.
            default: The default value to return if the key is not present.
        """
        res = self._data.get(key, default)
        if isinstance(res, dict):
            return HistoryDict(res)
        return res

    def to_dict(self) -> dict[str, dict | np.ndarray]:
        """
        Convert the HistoryDict to normal dictionary, with all nested HistoryDict instances.
        """
        return {k: self.__getitem__(k, wrap_dicts=False) for k in self._data.keys()}

    @classmethod
    def from_file(cls, fname: str) -> Self:
        """
        Create an HistoryDict from a text-file containing its serialization.

        Args:
            fname: The name of the file to read.
        """
        with open(fname) as f:
            data = orjson.loads(f.read())

        data = histdict_to_nparray(data)

        def _recompose(hist_dict):
            # hist_dict = {k: np.array(v) for k, v in hist_dict.items()}
            for _, checker, reconstructor in DESERIALIZATION_REGISTRY:
                if checker(hist_dict):
                    return reconstructor(hist_dict)
            raise ValueError("No matching type found for the given dictionary.")

        def _is_leaf(dic):
            if isinstance(dic, dict):
                if "iters" in dic:
                    return True
            return False

        return cls(jax.tree.map(_recompose, data, is_leaf=_is_leaf))

    def _ipython_key_completions_(self):
        return self._data.keys()


# Loading
# A registry to map types to their precedence, checker, and reconstructor functions
DESERIALIZATION_REGISTRY = []


def register_historydict_deserialization_fun(
    checker: Callable,
    reconstructor: Callable,
    precedence: int = 0,
):
    """
    Register a type with a precedence, checker, and reconstructor.

    Args:
        checker: A callable that takes a dictionary and returns True if it matches the type.
        reconstructor: A callable that takes a dictionary and reconstructs the object.
        precedence: An integer indicating the precedence of this type (higher is checked first).
    """
    DESERIALIZATION_REGISTRY.append((precedence, checker, reconstructor))
    # Sort the registry by precedence in descending order
    DESERIALIZATION_REGISTRY.sort(key=lambda x: x[0], reverse=True)


# Checker and reconstructor for History
def is_history(hist_dict):
    return "iters" in hist_dict


def reconstruct_history(hist_dict):
    from netket.utils.history.history import History

    if "Mean" in hist_dict:
        return History(hist_dict, main_value_name="Mean")
    return History(hist_dict)


# Register History with precedence 10
register_historydict_deserialization_fun(
    is_history, reconstruct_history, precedence=-10
)


"""
Json serializer for complex numbers.

Json does not support complex numbers, which are stored as {"real": ..., "imag": ...} dictionaries.
This is a workaround to store complex numbers in json files.
"""


def _is_number_list(x):
    return (
        isinstance(x, list) and len(x) > 0 and isinstance(x[0], (int, float, complex))
    )


def _is_complex_leaf(subtree):
    # Print only if you want to debug
    # print("checking if subtree is leaf", subtree)
    return (
        isinstance(subtree, dict) and "real" in subtree and "imag" in subtree
    ) or _is_number_list(subtree)


def _convert_complex(subtree):
    if isinstance(subtree, dict) and "real" in subtree and "imag" in subtree:
        return np.array(subtree["real"]) + 1j * np.array(subtree["imag"])
    return np.array(subtree)


def histdict_to_nparray(hist_dict):
    """
    Convert the {'real': ..., 'imag': ...} dictionaries to complex numbers.
    """
    return jax.tree.map(_convert_complex, hist_dict, is_leaf=_is_complex_leaf)



from collections.abc import Callable
from functools import partial

from einops import rearrange

import jax
import jax.numpy as jnp
from jax.tree_util import tree_map

from jax.experimental.shard_map import shard_map
from jax.sharding import PartitionSpec as P, PositionalSharding

from netket import jax as nkjax
from netket import config
from netket.jax._jacobian.default_mode import JacobianMode
from netket.utils import timing
from netket.utils.types import Array
from netket.utils.version_check import module_version

from netket.jax import _ntk as nt


@timing.timed
@partial(
    jax.jit,
    static_argnames=(
        "log_psi",
        "solver_fn",
        "chunk_size",
        "mode",
    ),
)
def srt_onthefly(
    log_psi,
    local_energies,
    parameters,
    model_state,
    samples,
    *,
    diag_shift: float | Array,
    solver_fn: Callable[[Array, Array], Array],
    mode: JacobianMode,
    proj_reg: float | Array | None = None,
    momentum: float | Array | None = None,
    old_updates: Array | None = None,
    chunk_size: int | None = None,
):
    N_mc = local_energies.size

    # Split all parameters into real and imaginary parts separately
    parameters_real, rss = nkjax.tree_to_real(parameters)

    # complex: (Nmc) -> (Nmc,2) - splitting real and imaginary output like 2 classes
    # real:    (Nmc) -> (Nmc,)  - no splitting
    def _apply_fn(parameters_real, samples):
        variables = {"params": rss(parameters_real), **model_state}
        log_amp = log_psi(variables, samples)

        if mode == "complex":
            re, im = log_amp.real, log_amp.imag
            return jnp.concatenate(
                (re[:, None], im[:, None]), axis=-1
            )  # shape [N_mc,2]
        else:
            return log_amp.real  # shape [N_mc, ]

    def jvp_f_chunk(parameters, vector, samples):
        r"""
        Creates the jvp of the function `_apply_fn` with respect to the parameters.
        This jvp is then evaluated in chunks of `chunk_size` samples.
        """
        f = lambda params: _apply_fn(params, samples)
        _, acc = jax.jvp(f, (parameters,), (vector,))
        return acc

    # compute rhs of the linear system
    local_energies = local_energies.flatten()
    de = local_energies - jnp.mean(local_energies)

    # At the moment the final vjp is centered by centering the auxiliary vector a.
    # This is the same as centering the jacobian but may have larger variance.
    dv = 2.0 * de / jnp.sqrt(N_mc)  # shape [N_mc,]
    if mode == "complex":
        dv = jnp.stack([jnp.real(dv), jnp.imag(dv)], axis=-1)  # shape [N_mc,2]
    else:
        dv = jnp.real(dv)  # shape [N_mc,]

    if momentum is not None:
        if old_updates is None:
            old_updates = tree_map(jnp.zeros_like, parameters_real)
        else:
            acc = nkjax.apply_chunked(
                jvp_f_chunk, in_axes=(None, None, 0), chunk_size=chunk_size
            )(parameters_real, old_updates, samples)

            avg = jnp.mean(acc, axis=0)
            acc = (acc - avg) / jnp.sqrt(N_mc)
            dv -= momentum * acc

    if mode == "complex":
        dv = jax.lax.collapse(dv, 0, 2)  # shape [2*N_mc,] or [N_mc, ] if not complex

    # Collect all samples on all MPI ranks, those label the columns of the T matrix
    all_samples = samples
    if config.netket_experimental_sharding:
        samples = jax.lax.with_sharding_constraint(
            samples, PositionalSharding(jax.devices()).reshape(-1, 1)
        )
        all_samples = jax.lax.with_sharding_constraint(
            samples, PositionalSharding(jax.devices()).replicate()
        )

    _jacobian_contraction = nt.empirical_ntk_by_jacobian(
        f=_apply_fn,
        trace_axes=(),
        vmap_axes=0,
    )

    def jacobian_contraction(samples, all_samples, parameters_real):
        if chunk_size is None:
            # STRUCTURED_DERIVATIVES returns a complex array, but the imaginary part is zero
            # shape [N_mc/p.size, N_mc, 2, 2]
            return _jacobian_contraction(samples, all_samples, parameters_real).real
        else:
            _all_samples, _ = nkjax.chunk(all_samples, chunk_size=chunk_size)
            ntk_local = jax.lax.map(
                lambda batch_lattice: _jacobian_contraction(
                    samples, batch_lattice, parameters_real
                ).real,
                _all_samples,
            )
            if mode == "complex":
                return rearrange(ntk_local, "nbatches i j z w -> i (nbatches j) z w")
            else:
                return rearrange(ntk_local, "nbatches i j -> i (nbatches j)")

    # If we are sharding, use shard_map manually
    if config.netket_experimental_sharding:
        mesh = jax.make_mesh((jax.device_count(),), ("i",), devices=jax.devices())
        # SAMPLES, ALL_SAMPLES PARAMETERS_REAL
        in_specs = (P("i", None), P(), P())
        out_specs = P("i", None, None, None)

        # By default, I'm not sure whether the jacobian_contraction of NeuralTangents
        # Is correctly automatically sharded across devices. So we force it to be
        # sharded with shard map to be sure

        # check rep:
        check_rep = module_version("jax") < (0, 4, 38)
        # shard_map is broken between 0.4.38 and (as of 25 march 2025) 0.5.3.
        # We assume any version after 0.4.38 'has a bug' that shows up as
        # None is not Iterable
        # it's a bug in check_rep, so we disable it in this case
        jacobian_contraction = shard_map(
            jacobian_contraction,
            mesh=mesh,
            in_specs=in_specs,
            out_specs=out_specs,
            check_rep=check_rep,
        )

    # This disables the nkjax.sharding_decorator in here, which might appear
    # in the apply function inside.
    with nkjax.sharding._increase_SHARD_MAP_STACK_LEVEL():
        ntk_local = jacobian_contraction(samples, all_samples, parameters_real).real

    # shape [N_mc, N_mc, 2, 2] or [N_mc, N_mc]
    ntk = jax.lax.with_sharding_constraint(
        ntk_local,
        PositionalSharding(jax.devices()).replicate(),
    )
    if mode == "complex":
        # shape [2*N_mc, 2*N_mc] checked with direct calculation of J^T J
        ntk = rearrange(ntk, "i j z w -> (i z) (j w)")

    # Center the NTK by multiplying with a carefully designed matrix
    # shape [N_mc, N_mc] symmetric matrix
    delta = jnp.eye(N_mc) - 1 / N_mc
    if mode == "complex":
        # shape [2*N_mc, 2*N_mc]
        # Gets applied to the sub-blocks corresponding to the real part and imaginary part
        delta_conc = jnp.zeros((2 * N_mc, 2 * N_mc)).at[0::2, 0::2].set(delta)
        delta_conc = delta_conc.at[1::2, 1::2].set(delta)
        delta_conc = delta_conc.at[0::2, 1::2].set(0.0)
        delta_conc = delta_conc.at[1::2, 0::2].set(0.0)
    else:
        delta_conc = delta

    # shape [2*N_mc, 2*N_mc] centering the jacobian
    ntk = (delta_conc @ (ntk @ delta_conc)) / N_mc

    # add diag shift
    ntk_shifted = ntk + diag_shift * jnp.eye(ntk.shape[0])

    # add projection regularization
    if proj_reg is not None:
        ntk_shifted = ntk_shifted + proj_reg / N_mc

    # some solvers return a tuple, some others do not.
    aus_vector = solver_fn(ntk_shifted, dv)
    if isinstance(aus_vector, tuple):
        aus_vector, info = aus_vector
    else:
        info = {}

    if info is None:
        info = {}

    # Center the vector, equivalent to centering
    # The Jacobian
    aus_vector = aus_vector / jnp.sqrt(N_mc)
    aus_vector = delta_conc @ aus_vector

    # shape [N_mc,2]
    if mode == "complex":
        aus_vector = aus_vector.reshape(-1, 2)
    # shape [N_mc // p.size,2]
    aus_vector = jax.lax.with_sharding_constraint(
        aus_vector,
        PositionalSharding(jax.devices()).reshape(
            -1, *(1 for _ in range(aus_vector.ndim - 1))
        ),
    )

    # _, vjp_fun = jax.vjp(f, parameters_real)
    vjp_fun = nkjax.vjp_chunked(
        _apply_fn,
        parameters_real,
        samples,
        chunk_size=chunk_size,
        chunk_argnums=1,
        nondiff_argnums=1,
    )

    (updates,) = vjp_fun(aus_vector)  # pytree [N_params,]

    if momentum is not None:
        updates = tree_map(lambda x, y: x + momentum * y, updates, old_updates)
        old_updates = updates

    return rss(updates), old_updates, info



from netket.utils.history.history import History
from netket.utils.history.accum import (
    accum_histories,
    accum_in_tree,
    accum_histories_in_tree,
)
from netket.utils.history.history_dict import HistoryDict



from typing import Any
from collections.abc import Callable

import jax
from jax.flatten_util import ravel_pytree
import flax

from netket import jax as nkjax
from netket.optimizer.solver import cholesky
from netket.vqs import MCState, FullSumState, VariationalState
from netket.utils import timing, struct
from netket.utils.types import ScalarOrSchedule, Optimizer, Array, PyTree
from netket.jax._jacobian.default_mode import JacobianMode
from netket.operator import AbstractOperator
from netket import stats as nkstats

from netket.driver.abstract_variational_driver import (
    AbstractVariationalDriver,
)
from netket._src.ngd.sr_srt_common import sr, srt
from netket._src.ngd.srt_onthefly import srt_onthefly
from netket._src.operator.hpsi_utils import make_logpsi_U_afun
from netket.experimental.observable.infidelity.expect import get_local_estimator

ApplyFun = Callable[[PyTree, Array], Array]
KernelArgs = tuple[ApplyFun, PyTree, Array, tuple[Any, ...]]
KernelFun = Callable[[PyTree, Array], KernelArgs]
DeriativesArgs = tuple[ApplyFun, PyTree, PyTree, Array]


@jax.jit
def _flatten_samples(x):
    # return x.reshape(-1, x.shape[-1])
    return jax.lax.collapse(x, 0, x.ndim - 1)


class Infidelity_SR(AbstractVariationalDriver):
    r"""
    Infidelity minimization with respect to a target state :math:`|\Phi\rangle` (with possibly an operator :math:`U` such that :math:`|\Phi\rangle \equiv U|\Phi\rangle`)
    using Variational Monte Carlo (VMC) and **Stochastic Reconfiguration/Natural Gradient Descent**.
    The optimization is analogous to the one of :class:`netket.experimental.driver.VMC_SR` for ground state.
    The infidelity :math:`I` among the variational state :math:`|\Psi\rangle` and the target state :math:`|\Phi\rangle` corresponds to:

    .. math::

        I = 1 - \frac{|\langle\Psi|\Phi\rangle|^2 }{ \langle\Psi|\Psi\rangle \langle\Phi|\Phi\rangle } = 1 - \frac{\langle\Psi|\hat{I}_{op}|\Psi\rangle }{ \langle\Psi|\Psi\rangle },

    where:

    .. math::

        \hat{I}_{op} = \frac{|\Phi\rangle\langle\Phi|}{\langle\Phi|\Phi\rangle}

    is the projector onto the target state :math:`|\Phi\rangle` which corresponds to an effective Hamiltonian.
    In this case, the effective local energy is :math:`H^{loc}(x) = \frac{\Phi(x)}{\Psi(x)} \mathbb{E}_{y \sim |\Phi(y)|^2}\left[\frac{\Psi(y)}{\Phi(y)}\right]`.

    For details see `Sinibaldi et al. <https://quantum-journal.org/papers/q-2023-10-10-1131/>`_ and `Gravina et al. <https://quantum-journal.org/papers/q-2025-07-22-1803/>`_.
    """

    target_state: VariationalState
    "The target variational state :math:`|\\Phi\rangle`."

    operator: AbstractOperator = None
    "Operator :math:`U`."

    cv_coeff: float = -0.5
    r"""
    Optional control variate coefficient for variance reduction in Monte Carlo estimation
    (see `Sinibaldi et al. <https://quantum-journal.org/papers/q-2023-10-10-1131/>`).
    If None, no control variate is used. Default to the optimal value -0.5.
    """

    # Settings set by user
    diag_shift: ScalarOrSchedule = struct.field(serialize=False)
    r"""
    The diagonal shift :math:`\lambda` in the curvature matrix.

    This can be a scalar or a schedule. If it is a schedule, it should be
    a function that takes the current step as input and returns the value of the shift.
    """
    proj_reg: ScalarOrSchedule = struct.field(serialize=False)

    momentum: bool = struct.field(serialize=False, default=False)
    r"""
    Flag specifying whether to use momentum in the optimisation.

    If `True`, the optimizer will use momentum to accumulate previous updates
    following the SPRING optimizer from
    `G.Goldshlager, N.Abrahamsen and L.Lin <https://arxiv.org/abs/2401.10190>`_
    to accumulate previous updates for better approximation of the exact SR with
    no significant performance penalty.
    """

    _ham: AbstractOperator = struct.field(pytree_node=False, serialize=False)

    _mode: str = struct.field(serialize=False)
    _chunk_size_bwd: int | None = struct.field(serialize=False)
    _use_ntk: bool = struct.field(serialize=False)
    _on_the_fly: bool = struct.field(serialize=False)
    _linear_solver_fn: Any = struct.field(serialize=False)

    # Internal things cached
    _unravel_params_fn: Any = struct.field(serialize=False)

    # Serialized state
    _old_updates: PyTree = None
    _dp: PyTree = struct.field(serialize=False)
    info: Any | None = None
    """
    PyTree to pass on information from the solver,e.g, the quadratic model.
    """

    def __init__(
        self,
        target_state: VariationalState,
        optimizer: Optimizer,
        *,
        operator: AbstractOperator = None,
        diag_shift: ScalarOrSchedule,
        proj_reg: ScalarOrSchedule | None = None,
        momentum: ScalarOrSchedule | None = None,
        linear_solver_fn: Callable[[Array, Array], Array] = cholesky,
        variational_state: MCState = None,
        chunk_size_bwd: int | None = None,
        mode: JacobianMode | None = None,
        use_ntk: bool | None = None,
        on_the_fly: bool | None = None,
    ):
        r"""
        Initialize the driver with the given arguments.

        .. warning::

            The optimizer should be an instance of `optax.sgd`. Other optimizers, while they might work,
            will not *make mathematical sense* in the context of the SR/NGD optimization.

        Args:
            target_state: The target state :math:`|\Phi\rangle` that must be matched.
            optimizer: The optimizer to use for the parameter updates. To perform proper
                SR/NGD optimization this should be an instance of `optax.sgd`, but can be
                any other optimizer if you are brave.
            operator: The operator :math:`U`.
            variational_state: The variational state to optimize.
            diag_shift: The diagonal regularization parameter :math:`\lambda` for the QGT/NTK.
            proj_reg: The regularization parameter for the projection of the updates.
                (This usually is not very important and can be left to None)
            momentum: (SPRING, disabled by default, read above for details) a number between [0,1]
                that specifies the damping factor of
                the previous updates and works somewhat similarly to the beta parameter of ADAM.
                The maximum amplification of  the step size in SPRING is
                :math:`A(\mu)=1/\sqrt{1-μ^2}`
                Thus the  amplification is at most a factor of :math:`A(0.9)=2.3` or
                :math:`A(0.99)=7.1`. Values around ``momentum = 0.8`` empirically work well.
                (Defaults to None)
            linear_solver_fn: The linear solver function to use for the NGD solver.
            mode: The mode used to compute the jacobian of the variational state.
                Can be `'real'` or `'complex'`. Real can be used for real-valued wavefunctions
                with a sign, to truncate the arbitrary phase of the wavefunction. This leads
                to lower computational cost.
            on_the_fly: Whether to compute the QGT or NTK using lazy evaluation methods.
                This usually requires less memory. (Defaults to None, which will
                automatically chose the potentially best method).
            chunk_size_bwd: The number of rows of the NTK or of the Jacobian evaluated
                in a single sweep.
            use_ntk: Wheter to compute the updates using the Neural Tangent Kernel (NTK)
                instead of the Quantum Geometric Tensor (QGT), aka switching between
                SR and minSR. (Defaults to None, which will automatically choose the best
                method)
        """

        if operator is not None:

            def _logpsi_fun(apply_fun, variables, x, *args):
                variables_applyfun, O = flax.core.pop(variables, "operator")

                xp, mels = O.get_conn_padded(x)
                xp = xp.reshape(-1, x.shape[-1])
                logpsi_xp = apply_fun(variables_applyfun, xp, *args)
                logpsi_xp = logpsi_xp.reshape(mels.shape)

                return jax.scipy.special.logsumexp(
                    logpsi_xp.astype(complex), axis=-1, b=mels
                )

            logUpsi_fun, new_variables = make_logpsi_U_afun(
                target_state._apply_fun, operator, target_state.variables
            )

            if isinstance(target_state, MCState):
                self.target_state = MCState(
                    sampler=target_state.sampler,
                    apply_fun=logUpsi_fun,
                    n_samples=target_state.n_samples,
                    variables=new_variables,
                )
                self.target_state.sampler_state = target_state.sampler_state

            if isinstance(target_state, FullSumState):
                self.target_state = FullSumState(
                    hilbert=target_state.hilbert,
                    apply_fun=logUpsi_fun,
                    variables=new_variables,
                )
        else:
            self.target_state = target_state

        if isinstance(variational_state, FullSumState):
            raise TypeError(
                "NGD drivers do not support FullSumState. Please use 'standard' drivers with SR."
            )
        super().__init__(
            variational_state, optimizer, minimized_quantity_name="Infidelity"
        )

        if use_ntk is None:
            use_ntk = variational_state.n_parameters > variational_state.n_samples
            print("Automatic SR implementation choice: ", "NTK" if use_ntk else "QGT")

        if on_the_fly is None:
            if use_ntk:
                on_the_fly = True
            else:
                on_the_fly = False
        elif on_the_fly and not use_ntk:
            raise ValueError(
                """
                `onthefly` is only supported when `use_ntk=True`.

                We plan to support this mode for the standard NGD in the future.
                In the meantime, use a standard VMC+SR QGTOnTheFly.
                """
            )

        self.diag_shift = diag_shift
        self.proj_reg = proj_reg
        self.momentum = momentum

        self.chunk_size_bwd = chunk_size_bwd
        self._use_ntk = use_ntk
        self.mode = mode
        self._on_the_fly = on_the_fly

        self._linear_solver_fn = linear_solver_fn

        _, unravel_params_fn = ravel_pytree(self.state.parameters)
        self._unravel_params_fn = jax.jit(unravel_params_fn)

        self._old_updates: PyTree = None
        self._dp: PyTree = None

        # PyTree to pass on information from the solver, e.g, the quadratic model
        self.info = None

        params_structure = jax.tree_util.tree_map(
            lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), self.state.parameters
        )
        if not nkjax.tree_ishomogeneous(params_structure):
            raise ValueError(
                "SRt only supports neural networks with all real or all complex parameters. "
                "Hybrid structures are not yet supported (but we would welcome contributions. Get in touch with us!)"
            )

    @timing.timed
    def _forward_and_backward(self):
        self.state.reset()

        # Compute the local infidelity estimator and average Infidelity
        local_energies, local_energies_cv = get_local_estimator(
            self.state,
            self.target_state,
            self.cv_coeff,
        )
        self._loss_stats = nkstats.statistics(1 - local_energies_cv)

        # Extract the hyperparameters which might be iteration dependent
        diag_shift = self.diag_shift
        proj_reg = self.proj_reg
        momentum = self.momentum
        if callable(diag_shift):
            diag_shift = diag_shift(self.step_count)
        if callable(proj_reg):
            proj_reg = proj_reg(self.step_count)
        if callable(momentum):
            momentum = momentum(self.step_count)

        if self.use_ntk:
            if self.on_the_fly:
                compute_sr_update_fun = srt_onthefly
            else:
                compute_sr_update_fun = srt
        else:
            if self.on_the_fly:
                raise NotImplementedError
            else:
                compute_sr_update_fun = sr

        samples = _flatten_samples(self.state.samples)
        self._dp, self._old_updates, self.info = compute_sr_update_fun(
            self.state._apply_fun,
            local_energies,
            self.state.parameters,
            self.state.model_state,
            samples,
            diag_shift=diag_shift,
            solver_fn=self._linear_solver_fn,
            mode=self.mode,
            proj_reg=proj_reg,
            momentum=momentum,
            old_updates=self._old_updates,
            chunk_size=self.chunk_size_bwd,
        )

        self._dp = jax.tree_util.tree_map(lambda x: -x, self._dp)

        return self._dp

    @timing.timed
    def _log_additional_data(self, log_dict: dict, step: int):
        """
        Method to be implemented in sub-classes of AbstractVariationalDriver to
        log additional data at every step.
        This method is called at every iteration when executing with `run`.

        Args:
            log_dict: The dictionary containing all logged data. It must be
                **modified in-place** adding new keys.
            step: the current step number.
        """
        # Always log the acceptance.
        if hasattr(self.state, "sampler_state"):
            acceptance = getattr(self.state.sampler_state, "acceptance", None)
            if acceptance is not None:
                log_dict["acceptance"] = acceptance

        # Log the quadratic model if requested.
        if self.info is not None:
            log_dict["info"] = self.info

    @property
    def mode(self) -> JacobianMode:
        """
        The mode used to compute the jacobian of the variational state. Can be `'real'`, `'complex'`, or 'onthefly'.

        - `'real'` mode truncates imaginary part of the wavefunction, useful for real-valued wf with a sign.
        - `'complex'` is the general implementation that always works.
        - `onthefly` uses a lazy implementation of the neural tangent kernel and does not compute the jacobian.

        This internally uses :func:`netket.jax.jacobian`. See that function for a more complete documentation.
        """
        return self._mode

    @mode.setter
    def mode(self, mode: str | JacobianMode | None):
        # TODO: Add support for 'onthefly' mode
        # At the moment, onthefly is only supported for use_ntk=True.
        # We raise a warning if the user tries to use it with use_ntk=False.
        if mode is None:
            mode = nkjax.jacobian_default_mode(
                self.state._apply_fun,
                self.state.parameters,
                self.state.model_state,
                self.state.hilbert.random_state(jax.random.key(1), 3),
                warn=False,
            )

        # TODO: Add support for 'holomorphic' mode
        # At the moment we only support 'real' and 'complex' modes for jacobian.
        # We raise an error if the user tries to use 'holomorphic' mode.
        if mode not in ["complex", "real"]:
            raise ValueError(
                "`mode` only supports 'jacobian_real' for real-valued wavefunctions, and 'jacobian_complex' for complex valued wave functions."
                "`holomorphic` is not yet supported, but could be contributed in the future. \n"
                f"You gave {mode}"
            )

        self._mode = mode

    @property
    def on_the_fly(self) -> bool:
        """
        Whether
        """
        return self._on_the_fly

    @property
    def use_ntk(self) -> bool:
        r"""
        Whether to use the Neural Tangent Kernel (NTK) instead of the Quantum Geometric Tensor (QGT) to compute the update.
        """
        return self._use_ntk

    @property
    def chunk_size_bwd(self) -> int:
        """
        Chunk size for backward-mode differentiation. This reduces memory pressure at a potential cost of higher computation time.

        If computing the jacobian, the jacobian is computed in blocks of `chunk_size_bwd` rows.
        If computing the NTK lazily, this is the number of rows of NTK evaluated in a single sweep.
        The chunk size does not affect the result, up to numerical precision.
        """
        return self._chunk_size_bwd

    @chunk_size_bwd.setter
    def chunk_size_bwd(self, value: int | None):
        if not isinstance(value, int | None):
            raise TypeError("chunk_size must be an integer or None")
        self._chunk_size_bwd = value



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .tdvp import TDVP
from .tdvp_schmitt import TDVPSchmitt
from .vmc_sr import VMC_SR, VMC_SRt as _VMC_SRt_deprecated
from .infidelity_sr import Infidelity_SR

_deprecations = {
    # May 2024, NetKet 3.12
    "VMC_SRt": (
        "netket.experimental.driver.VMC_SRt is deprecated: use the new SR driver "
        "netket.experimental.driver.VMC_SR(..., use_ntk=True) (netket >= 3.17)",
        _VMC_SRt_deprecated,
    ),
}

from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr
from netket.utils import _hide_submodules

__getattr__ = _deprecation_getattr(__name__, _deprecations)
_hide_submodules(__name__)

del _deprecation_getattr



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ["FermionOperator2nd", "FermionOperator2ndJax"]

from .base import FermionOperator2ndBase
from .numba import FermionOperator2ndNumba
from .jax import FermionOperator2ndJax

FermionOperator2nd = FermionOperator2ndJax



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

from flax.linen import Module

import jax.numpy as jnp


from netket.utils.partial import HashablePartial
from netket.utils.types import ModuleOrApplyFun, PyTree, Array

from netket.utils import struct


def get_afun_if_module(mod_or_fun) -> Callable:
    """Returns the apply function if it's a module. Does nothing otherwise."""
    if hasattr(mod_or_fun, "apply"):
        return mod_or_fun.apply
    else:
        return mod_or_fun


@struct.dataclass
class WrappedApplyFun:
    """Wraps a callable to be a module-like object with the method `apply`."""

    apply: Callable[[PyTree, Array], Array]
    """The wrapped callable."""

    def __repr__(self):
        return f"{type(self).__name__}(apply={self.apply}, hash={hash(self)})"


def wrap_afun(mod_or_fun: ModuleOrApplyFun) -> Module:
    """Wraps a callable to be a module-like object with the method `apply`.
    Does nothing if it already has an apply method.
    """
    if hasattr(mod_or_fun, "apply"):
        return mod_or_fun  # type: ignore
    else:
        return WrappedApplyFun(mod_or_fun)  # type: ignore[return-type]


def wrap_to_support_scalar(fun):
    """
    Wraps the flax-compatible apply function, assuming that the state input is the
    second argument, so that it always calls the wrapped function with a tensor with at
    least 2 dimensions.

    If the input was 1-dimensional, returns a scalar instead of a vector with 1 element.

    DEVNOTE: This function is used because some parts of NetKet make use of the fact
    that when we call the logψ with a single bitstring, we get a scalar out. This is
    useful when calling `jax.grad(logψ)`.
    This also makes sure that users only need to write networks that work on batches,
    and not necessarily on scalars.

    Args:
        fun: A flax-compatible function.

    Returns:
        A wrapped function, returned as an `HashablePartial` in order not to retrigger
        compilation.
    """

    def maybe_scalar_fun(apply_fun, pars, x, *args, **kwargs):
        xb = jnp.atleast_2d(x)
        res = apply_fun(pars, xb, *args, **kwargs)
        # support models with mutable state
        if isinstance(res, tuple):
            res_val = res[0]
            res_val = res_val.reshape(()) if x.ndim == 1 else res_val
            res = (res_val, res[1])
        else:
            res = res.reshape(()) if x.ndim == 1 else res
        return res

    return HashablePartial(maybe_scalar_fun, fun)



import jax
import flax

from netket import jax as nkjax
from netket.operator import ContinuousOperator


def make_logpsi_U_afun(logpsi_fun, U, variables):
    """Wraps an apply_fun into another one that multiplies it by an Unitary transformation U.
    This wrapper is made such that the Unitary is passed as the model_state
    of the new wrapped function, and therefore changes to the angles/coefficients
    of the Unitary should not trigger recompilation.
    Args:
        logpsi_fun: a function that takes as input variables and samples
        U: a {class}`nk.operator.JaxDiscreteOperator`
        variables: The variables used to call *logpsi_fun*
    Returns:
        A tuple, where the first element is a new function with the same signature as
        the original **logpsi_fun** and a set of new variables to be used to call it.
    """
    # wrap apply_fun into logpsi logpsi_U
    logpsiU_fun = nkjax.HashablePartial(_logpsi_U_fun, logpsi_fun)

    # Insert a new 'model_state' key to store the Unitary. This only works
    # if U is a pytree that can be flattened/unflattened.
    new_variables = flax.core.copy(variables, {"operator": U})

    return logpsiU_fun, new_variables


def _logpsi_U_fun(apply_fun, variables, x, *args):
    """
    This should be used as a wrapper to the original apply function, adding
    to the `variables` dictionary (in model_state) a new key `unitary` with
    a jax-compatible operator.
    """
    variables_applyfun, U = flax.core.pop(variables, "operator")

    if isinstance(U, ContinuousOperator):
        res = U._expect_kernel(apply_fun, variables_applyfun, x)
    else:
        xp, mels = U.get_conn_padded(x)
        xp = xp.reshape(-1, x.shape[-1])
        logpsi_xp = apply_fun(variables_applyfun, xp, *args)
        logpsi_xp = logpsi_xp.reshape(mels.shape).astype(complex)

        res = jax.scipy.special.logsumexp(logpsi_xp, axis=-1, b=mels)
    return res



from functools import partial

import jax
import jax.numpy as jnp

from netket.vqs import FullSumState, expect, expect_and_grad
import netket.jax as nkjax
from netket.stats import Stats

from netket.experimental.observable.infidelity.infidelity_operator import (
    InfidelityOperator,
)


@expect.dispatch
def infidelity(vstate: FullSumState, op: InfidelityOperator):
    if op.hilbert != vstate.hilbert:
        raise TypeError("Hilbert spaces should match")

    return infidelity_fullsum_inner(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        op.target_state.to_array(),
        vstate._all_states,
        return_grad=False,
    )


@expect_and_grad.dispatch
def infidelity(  # noqa: F811
    vstate: FullSumState,
    op: InfidelityOperator,
    *,
    mutable,
):
    if op.hilbert != vstate.hilbert:
        raise TypeError("Hilbert spaces should match")

    return infidelity_fullsum_inner(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        op.target_state.to_array(),
        vstate._all_states,
        return_grad=True,
    )


@partial(jax.jit, static_argnames=("afun", "return_grad"))
def infidelity_fullsum_inner(
    afun,
    params,
    model_state,
    state_t,
    σ,
    return_grad,
):
    def expect_fun(params, σ):
        W = {"params": params, **model_state}

        psi = jnp.exp(afun(W, σ))
        psi /= jnp.linalg.norm(psi)

        I = 1 - jnp.absolute(jnp.vdot(psi.conj(), state_t)) ** 2

        return I

    if not return_grad:
        I = expect_fun(params, σ)

        I_stats = Stats(
            mean=I,
            error_of_mean=0.0,
            variance=0.0,
        )

        return I_stats

    I, I_vjp_fun = nkjax.vjp(expect_fun, params, σ, conjugate=True)

    I_grad = I_vjp_fun(jnp.ones_like(I))[0]

    I_stats = Stats(
        mean=I,
        error_of_mean=0.0,
        variance=0.0,
    )

    return I_stats, I_grad



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .base import MetropolisRule


class FixedRule(MetropolisRule):
    r"""
    A transition rule relaxing and doing nothing.

    You can use it to make a CombinedRule not act on a certain subspace at all.
    """

    def transition(rule, sampler, machine, parameters, state, key, σ):
        return σ, None

    def __repr__(self):
        return "FixedRule()"



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

import types


mpi4py_available = False
mpi4jax_available = False
MPI_py_comm = None  # type: ignore
MPI_jax_comm = None  # type: ignore
n_nodes = 1
node_number = 0
rank = 0

FakeMPI = types.ModuleType("FakeMPI", "FakeMPI Module")
FakeMPI.COMM_WORLD = None

MPI = FakeMPI



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Callable
from collections.abc import Hashable

import jax
import jax.numpy as jnp

from netket.utils.types import DType, PyTree, Array

from netket.hilbert import AbstractHilbert
from netket.operator import ContinuousOperator
from netket.utils import HashableArray, struct


class PotentialEnergy(ContinuousOperator):
    r"""Returns the local potential energy defined in afun"""

    _potential_fun: Callable = struct.static_field()
    _coefficient: Array

    def __init__(
        self,
        hilbert: AbstractHilbert,
        afun: Callable,
        coefficient: float = 1.0,
        dtype: DType | None = None,
    ):
        r"""
        Args:
            hilbert: The underlying Hilbert space on which the operator is defined
            afun: The potential energy as function of x
            coefficient: A coefficient for the ContinuousOperator object
            dtype: Data type of the coefficient
        """

        self._potential_fun = afun
        self._coefficient = jnp.array(coefficient, dtype=dtype)

        super().__init__(hilbert, self._coefficient.dtype)

    @property
    def coefficient(self) -> Array:
        return self._coefficient

    @property
    def is_hermitian(self) -> bool:
        return True

    def _expect_kernel(
        self,
        logpsi: Callable,
        params: PyTree,
        x: Array,
    ) -> Array:
        x_dtype = jnp.promote_types(self.dtype, x.dtype)
        return self.coefficient * jax.vmap(self._potential_fun, in_axes=(0,))(
            x.astype(x_dtype)
        )

    @struct.property_cached(pytree_ignore=True)
    def _attrs(self) -> tuple[Hashable, ...]:
        return (
            self.hilbert,
            self._potential_fun,
            self.dtype,
            HashableArray(self.coefficient),
        )

    def __repr__(self):
        return (
            f"Potential(coefficient={self.coefficient}, function={self._potential_fun})"
        )



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import rules


# Deprecated bindings from stabilisaation
from netket.sampler import (
    MetropolisFermionHop as _deprecated_MetropolisParticleExchange,
)

_deprecations = {
    # June 2024, NetKet 3.13
    "MetropolisParticleExchange": (
        "netket.experimental.sampler.MetropolisParticleExchange is deprecated: use "
        "netket.sampler.MetropolisFermionHop (netket >= 3.13)",
        _deprecated_MetropolisParticleExchange,
    ),
}


from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr
from netket.utils import _hide_submodules

__getattr__ = _deprecation_getattr(__name__, _deprecations)
_hide_submodules(__name__)

del _deprecation_getattr



from typing import Any
from collections.abc import Callable

import jax
from jax.flatten_util import ravel_pytree

from netket import jax as nkjax
from netket.optimizer.solver import cholesky
from netket.vqs import MCState, FullSumState
from netket.utils import timing, struct
from netket.utils.types import ScalarOrSchedule, Optimizer, Array, PyTree
from netket.jax._jacobian.default_mode import JacobianMode
from netket.operator import AbstractOperator
from netket import stats as nkstats

from netket.driver.abstract_variational_driver import (
    AbstractVariationalDriver,
)
from netket._src.ngd.sr_srt_common import sr, srt
from netket._src.ngd.srt_onthefly import srt_onthefly


ApplyFun = Callable[[PyTree, Array], Array]
KernelArgs = tuple[ApplyFun, PyTree, Array, tuple[Any, ...]]
KernelFun = Callable[[PyTree, Array], KernelArgs]
DeriativesArgs = tuple[ApplyFun, PyTree, PyTree, Array]


@jax.jit
def _flatten_samples(x):
    # return x.reshape(-1, x.shape[-1])
    return jax.lax.collapse(x, 0, x.ndim - 1)


def VMC_SRt(
    hamiltonian: AbstractOperator,
    optimizer: Optimizer,
    *,
    diag_shift: ScalarOrSchedule,
    linear_solver_fn: Callable[[jax.Array, jax.Array], jax.Array] = cholesky,
    mode: str | None = None,
    jacobian_mode: str | None = None,
    variational_state: MCState = None,
):
    if mode is None:
        mode = jacobian_mode
    elif mode is not None and jacobian_mode is not None:
        raise ValueError(
            "`jacobian_mode` is deprecated and renamed to `mode`. Just declare `mode`."
        )

    return VMC_SR(
        hamiltonian,
        optimizer,
        diag_shift=diag_shift,
        linear_solver_fn=linear_solver_fn,
        mode=mode,
        variational_state=variational_state,
        use_ntk=True,
        on_the_fly=False,
    )


class VMC_SR(AbstractVariationalDriver):
    r"""
    Energy minimization using Variational Monte Carlo (VMC) and **Stochastic Reconfiguration/Natural Gradient Descent**.
    This driver is mathematically equivalent to the standard :class:`netket.driver.VMC` with the
    preconditioner :class:`netket.optimizer.SR(solver=netket.optimizer.solvers.cholesky) <netket.optimizer.SR>`,
    but can easily switch between the standard and the kernel/minSR formulation of Natural Gradient Descent.

    - The standard formulation computes the updates as:

    .. math::
        \delta \theta = \tau (X^TX + \lambda \mathbb{I}_{N_P})^{-1} X^T E^{loc},

    where :math:`X \in R^{N_s \times N_p}` is the Jacobian of the log-wavefunction, with :math:`N_p` the number of parameters
    and :math:`N_s` the number of samples. The vector :math:`E^{loc}` is the centered local estimator for the local energies.

    - The kernel/minSR formulation computes the updates as:

    .. math::
        \delta \theta = \tau X^T(XX^T + \lambda \mathbb{I}_{2N_s})^{-1} E^{loc},

    The regularization parameter :math:`\lambda` is the `diag_shift` parameter of the driver, which can be
    a scalar or a schedule.
    The updates are then applied to the parameters using the `optimizer` which in general should be `optax.sgd`.


    Matrix Inversion
    ----------------

    The matrix inversion of both methods is performed using a linear solver, which can be specified by the user.
    This must be a function, the :code:`linear_solver_fun` argument, which has the following signature:

    .. code-block:: python

        linear_solver_fn(A: Matrix, b: vector) -> tuple[jax.Array[vector], dict]

    Where the vector is the solution and the dictionary may contain additional information about the solver or be None.
    The standard solver is based on the Cholesky decomposition :func:`~netket.optimizer.solver.cholesky`, but any other
    solver from `JAX <https://jax.readthedocs.io/en/latest/jax.experimental.linalg.html>`_, `netket solvers <dense-solvers>`_ or a
    custom-written one can be used.


    Natural Gradient Descent
    ------------------------

    Stochastic Reconfiguration is equivalent to the Natural Gradient Descent method introduced by
    `Amari 1998 <https://ieeexplore.ieee.org/abstract/document/6790500/>`_ in the context of neural network training,
    assuming that the *natural metric* of the space of wave-functions is the Fubini-Study metric. This was first
    studied by `Stokes et Al 2019 <https://arxiv.org/abs/1909.02108>`_ and called
    *quantum Natural Gradient Descent*.

    While stochastic reconfiguration has been heavily studied in the context of VMC, there is a vast literature
    in the Machine Learning community on the use of NGD, and tuning carefully the diag shift and the learning rate.

    A very good introduction to the mathematics of Information Geometry and NGD is found in
    `Bai et Al <https://arxiv.org/pdf/2202.06232>`_ and further studied in `Shrestha et Al 2022 <https://arxiv.org/pdf/2303.05473>`_.
    From the Physicist point of view, a good discussion on the choice of the metric function (QGT vs Fisher Matrix)
    is found in `Stokes et Al 2022 <https://arxiv.org/pdf/2203.14824>`_ (section 4 in particoular).
    For a comprehensive review of the method, we suggest the review by
    `Martens 2014 <https://arxiv.org/abs/1412.1193>`_.


    Momentum / SPRING
    -----------------
    When `momentum` is used, this driver implements the SPRING optimizer in
    `Goldshlager et Al. (2024) <https://arxiv.org/abs/2401.10190>`_
    to accumulate previous updates for better approximation of the exact SR with
    no significant performance penalty.

    `momentum` μ is a number between [0,1] that specifies the damping factor of
    the previous updates and works somewhat similarly to the beta parameter of ADAM.
    The difference is that rather than simply adding the damped previous update to the
    new update, SPRING uses the damped previous update to fill in the components of the
    SR direction that are not sampled by the current batch of walkers, resulting in a
    more accurate and less noisy estimate. Since SPRING only uses the previous update to
    fill in directions that are orthogonal to the current one, the maximum amplification of
    the step size in SPRING is :math:`A(\mu) = 1/\sqrt{1-μ^2}` rather than :math:`1/(1-μ)`.

    Thus the  amplification is at most a factor of :math:`A(0.9)=2.3` or
    :math:`A(0.99)=7.1`.
    ** Values that empirically work are around 0.8. **

    Some progress has been made on theoretically analyzing this parameter, in particular
    `Section 3 of Epperly et Al. <https://arxiv.org/pdf/2411.19877>`_ demonstrates (albeit
    in a significantly  simplified linear least-squares setting) that SPRING can be interpreted
    as iteratively estimating a regularized SR direction, with the amount of regularization
    proportional to  the value of 1-momentum. Additional insights regarding the behavior of
    some SPRING-like algorithms, albeit still in the linear least-squares setting, are presented in
    `Goldshlager et Al. (2025) <https://arxiv.org/pdf/2502.00882>`_ .


    Implementation details
    ------------------------
    The kernel-trick/NTK based implementation can run with both a direct calculation of the jacobian
    (`on_the_fly=False`) or with a lazy evaluation of the NTK (`on_the_fly=True`). The latter is more
    computationally efficient for networks that reuse the parameters many times for every forward pass
    (convolutions, attention layers, but not dense layers...) and generally uses less memory.

    However, the on the fly implementation relies on some JAX compiler behaviour, so it might at times
    have worse performance. We suggest you check on your specific model. For a more detailed explanation
    of the on-the-fly implementation of the NTK, we refer to `Novak et Al 2022 <https://arxiv.org/pdf/2206.08720>`_.
    The algorithm netket uses is the layer-wise jacobian contraction method (sec 3.2) of the manuscript.

    The default choice is to use the ``on_the_fly=True`` mode.

    References
    ----------
    - Stochastic Reconfiguration was originally introduced in the QMC field by `Sorella <https://arxiv.org/abs/cond-mat/9803107>`_.
      The method was later shown to be equivalent to the Natural Gradient Descent method introduced by
      `Amari <https://ieeexplore.ieee.org/abstract/document/6790500/>`_ for the Fubini-Study metric.

    - The *kernel trick* which makes NGD/SR feasible in the large-parameter count limit was originally introduced
      to the field of NQS by `Chen & Heyl <https://arxiv.org/abs/2302.01941>`_ under the name of `minSR`.
      `Rende & Al <https://arxiv.org/abs/2310.05715>`_ proposed a simpler derivation in terms of the Kernel trick.

      It's interesting to note that those tricks were first mentioned by `Ren & Goldfarb <https://arxiv.org/abs/1906.02353>`_
      in the ML community.

    - When using Momentum you should cite `G.Goldshlager et Al. (2024) <https://arxiv.org/abs/2401.10190>`_.
    """

    # Settings set by user
    diag_shift: ScalarOrSchedule = struct.field(serialize=False)
    r"""
    The diagonal shift :math:`\lambda` in the curvature matrix.

    This can be a scalar or a schedule. If it is a schedule, it should be
    a function that takes the current step as input and returns the value of the shift.
    """
    proj_reg: ScalarOrSchedule = struct.field(serialize=False)

    momentum: bool = struct.field(serialize=False, default=False)
    r"""
    Flag specifying whether to use momentum in the optimisation.

    If `True`, the optimizer will use momentum to accumulate previous updates
    following the SPRING optimizer from
    `G.Goldshlager, N.Abrahamsen and L.Lin <https://arxiv.org/abs/2401.10190>`_
    to accumulate previous updates for better approximation of the exact SR with
    no significant performance penalty.
    """

    _ham: AbstractOperator = struct.field(pytree_node=False, serialize=False)

    _mode: str = struct.field(serialize=False)
    _chunk_size_bwd: int | None = struct.field(serialize=False)
    _use_ntk: bool = struct.field(serialize=False)
    _on_the_fly: bool = struct.field(serialize=False)
    _linear_solver_fn: Any = struct.field(serialize=False)

    # Internal things cached
    _unravel_params_fn: Any = struct.field(serialize=False)

    # Serialized state
    _old_updates: PyTree = None
    _dp: PyTree = struct.field(serialize=False)
    info: Any | None = None
    """
    PyTree to pass on information from the solver,e.g, the quadratic model.
    """

    def __init__(
        self,
        hamiltonian: AbstractOperator,
        optimizer: Optimizer,
        *,
        diag_shift: ScalarOrSchedule,
        proj_reg: ScalarOrSchedule | None = None,
        momentum: ScalarOrSchedule | None = None,
        linear_solver_fn: Callable[[Array, Array], Array] = cholesky,
        variational_state: MCState = None,
        chunk_size_bwd: int | None = None,
        mode: JacobianMode | None = None,
        use_ntk: bool | None = None,
        on_the_fly: bool | None = None,
    ):
        r"""
        Initialize the driver with the given arguments.

        .. warning::

            The optimizer should be an instance of `optax.sgd`. Other optimizers, while they might work,
            will not *make mathematical sense* in the context of the SR/NGD optimization.

        Args:
            hamiltonian: The Hamiltonian of which the ground-state is to be found.
            optimizer: The optimizer to use for the parameter updates. To perform proper
                SR/NGD optimization this should be an instance of `optax.sgd`, but can be
                any other optimizer if you are brave.
            variational_state: The variational state to optimize.
            diag_shift: The diagonal regularization parameter :math:`\lambda` for the QGT/NTK.
            proj_reg: The regularization parameter for the projection of the updates.
                (This usually is not very important and can be left to None)
            momentum: (SPRING, disabled by default, read above for details) a number between [0,1]
                that specifies the damping factor of
                the previous updates and works somewhat similarly to the beta parameter of ADAM.
                The maximum amplification of  the step size in SPRING is
                :math:`A(\mu)=1/\sqrt{1-μ^2}`
                Thus the  amplification is at most a factor of :math:`A(0.9)=2.3` or
                :math:`A(0.99)=7.1`. Values around ``momentum = 0.8`` empirically work well.
                (Defaults to None)
            linear_solver_fn: The linear solver function to use for the NGD solver.
            mode: The mode used to compute the jacobian of the variational state.
                Can be `'real'` or `'complex'`. Real can be used for real-valued wavefunctions
                with a sign, to truncate the arbitrary phase of the wavefunction. This leads
                to lower computational cost.
            on_the_fly: Whether to compute the QGT or NTK using lazy evaluation methods.
                This usually requires less memory. (Defaults to None, which will
                automatically chose the potentially best method).
            chunk_size_bwd: The number of rows of the NTK or of the Jacobian evaluated
                in a single sweep.
            use_ntk: Wheter to compute the updates using the Neural Tangent Kernel (NTK)
                instead of the Quantum Geometric Tensor (QGT), aka switching between
                SR and minSR. (Defaults to None, which will automatically choose the best
                method)
        """
        if isinstance(variational_state, FullSumState):
            raise TypeError(
                "NGD drivers do not support FullSumState. Please use 'standard' drivers with SR."
            )
        super().__init__(variational_state, optimizer, minimized_quantity_name="Energy")

        if use_ntk is None:
            use_ntk = variational_state.n_parameters > variational_state.n_samples
            print("Automatic SR implementation choice: ", "NTK" if use_ntk else "QGT")

        if on_the_fly is None:
            if use_ntk:
                on_the_fly = True
            else:
                on_the_fly = False
        elif on_the_fly and not use_ntk:
            raise ValueError(
                """
                `onthefly` is only supported when `use_ntk=True`.

                We plan to support this mode for the standard NGD in the future.
                In the meantime, use a standard VMC+SR QGTOnTheFly.
                """
            )

        self._ham = hamiltonian

        self.diag_shift = diag_shift
        self.proj_reg = proj_reg
        self.momentum = momentum

        self.chunk_size_bwd = chunk_size_bwd
        self._use_ntk = use_ntk
        self.mode = mode
        self._on_the_fly = on_the_fly

        self._linear_solver_fn = linear_solver_fn

        _, unravel_params_fn = ravel_pytree(self.state.parameters)
        self._unravel_params_fn = jax.jit(unravel_params_fn)

        self._old_updates: PyTree = None
        self._dp: PyTree = None

        # PyTree to pass on information from the solver, e.g, the quadratic model
        self.info = None

        params_structure = jax.tree_util.tree_map(
            lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), self.state.parameters
        )
        if not nkjax.tree_ishomogeneous(params_structure):
            raise ValueError(
                "SRt only supports neural networks with all real or all complex parameters. "
                "Hybrid structures are not yet supported (but we would welcome contributions. Get in touch with us!)"
            )

    @timing.timed
    def _forward_and_backward(self):
        self.state.reset()

        # Compute the local energy estimator and average Energy
        local_energies = self.state.local_estimators(self._ham)
        self._loss_stats = nkstats.statistics(local_energies)

        # Extract the hyperparameters which might be iteration dependent
        diag_shift = self.diag_shift
        proj_reg = self.proj_reg
        momentum = self.momentum
        if callable(diag_shift):
            diag_shift = diag_shift(self.step_count)
        if callable(proj_reg):
            proj_reg = proj_reg(self.step_count)
        if callable(momentum):
            momentum = momentum(self.step_count)

        if self.use_ntk:
            if self.on_the_fly:
                compute_sr_update_fun = srt_onthefly
            else:
                compute_sr_update_fun = srt
        else:
            if self.on_the_fly:
                raise NotImplementedError
            else:
                compute_sr_update_fun = sr

        samples = _flatten_samples(self.state.samples)
        self._dp, self._old_updates, self.info = compute_sr_update_fun(
            self.state._apply_fun,
            local_energies,
            self.state.parameters,
            self.state.model_state,
            samples,
            diag_shift=diag_shift,
            solver_fn=self._linear_solver_fn,
            mode=self.mode,
            proj_reg=proj_reg,
            momentum=momentum,
            old_updates=self._old_updates,
            chunk_size=self.chunk_size_bwd,
        )

        return self._dp

    @timing.timed
    def _log_additional_data(self, log_dict: dict, step: int):
        """
        Method to be implemented in sub-classes of AbstractVariationalDriver to
        log additional data at every step.
        This method is called at every iteration when executing with `run`.

        Args:
            log_dict: The dictionary containing all logged data. It must be
                **modified in-place** adding new keys.
            step: the current step number.
        """
        # Always log the acceptance.
        if hasattr(self.state, "sampler_state"):
            acceptance = getattr(self.state.sampler_state, "acceptance", None)
            if acceptance is not None:
                log_dict["acceptance"] = acceptance

        # Log the quadratic model if requested.
        if self.info is not None:
            log_dict["info"] = self.info

    @property
    def mode(self) -> JacobianMode:
        """
        The mode used to compute the jacobian of the variational state. Can be `'real'`, `'complex'`, or 'onthefly'.

        - `'real'` mode truncates imaginary part of the wavefunction, useful for real-valued wf with a sign.
        - `'complex'` is the general implementation that always works.
        - `onthefly` uses a lazy implementation of the neural tangent kernel and does not compute the jacobian.

        This internally uses :func:`netket.jax.jacobian`. See that function for a more complete documentation.
        """
        return self._mode

    @mode.setter
    def mode(self, mode: str | JacobianMode | None):
        # TODO: Add support for 'onthefly' mode
        # At the moment, onthefly is only supported for use_ntk=True.
        # We raise a warning if the user tries to use it with use_ntk=False.
        if mode is None:
            mode = nkjax.jacobian_default_mode(
                self.state._apply_fun,
                self.state.parameters,
                self.state.model_state,
                self.state.hilbert.random_state(jax.random.key(1), 3),
                warn=False,
            )

        # TODO: Add support for 'holomorphic' mode
        # At the moment we only support 'real' and 'complex' modes for jacobian.
        # We raise an error if the user tries to use 'holomorphic' mode.
        if mode not in ["complex", "real"]:
            raise ValueError(
                "`mode` only supports 'jacobian_real' for real-valued wavefunctions, and 'jacobian_complex' for complex valued wave functions."
                "`holomorphic` is not yet supported, but could be contributed in the future. \n"
                f"You gave {mode}"
            )

        self._mode = mode

    @property
    def on_the_fly(self) -> bool:
        """
        Whether
        """
        return self._on_the_fly

    @property
    def use_ntk(self) -> bool:
        r"""
        Whether to use the Neural Tangent Kernel (NTK) instead of the Quantum Geometric Tensor (QGT) to compute the update.
        """
        return self._use_ntk

    @property
    def chunk_size_bwd(self) -> int:
        """
        Chunk size for backward-mode differentiation. This reduces memory pressure at a potential cost of higher computation time.

        If computing the jacobian, the jacobian is computed in blocks of `chunk_size_bwd` rows.
        If computing the NTK lazily, this is the number of rows of NTK evaluated in a single sweep.
        The chunk size does not affect the result, up to numerical precision.
        """
        return self._chunk_size_bwd

    @chunk_size_bwd.setter
    def chunk_size_bwd(self, value: int | None):
        if not isinstance(value, int | None):
            raise TypeError("chunk_size must be an integer or None")
        self._chunk_size_bwd = value



from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp
from jax.sharding import PositionalSharding

from netket import jax as nkjax
from netket import config
from netket.utils import timing
from netket.utils.types import Array


@timing.timed
@partial(
    jax.jit,
    static_argnames=(
        "solver_fn",
        "mode",
    ),
)
def _compute_srt_update(
    O_L,
    dv,
    *,
    diag_shift: float | Array,
    solver_fn: Callable[[Array, Array], Array],
    mode: str,
    proj_reg: float | Array | None = None,
    momentum: float | Array | None = None,
    old_updates: Array | None = None,
    params_structure,
):
    if momentum is not None:
        dv -= momentum * (O_L @ old_updates)

    # (#ns, np) -> (ns, #np)
    O_LT = O_L
    if config.netket_experimental_sharding:
        nkjax.sharding.pad_axis_for_sharding(O_LT, axis=1, padding_value=0.0)
        O_LT = jax.lax.with_sharding_constraint(
            O_LT,
            PositionalSharding(jax.devices()).reshape(1, -1),
        )
        dv = jax.lax.with_sharding_constraint(
            dv, PositionalSharding(jax.devices()).replicate()
        )

    # This does the contraction (ns, #np) x (#np, ns) -> (ns, ns).
    # When using sharding the sum over #ns is done automatically.
    # When using MPI we need to do it manually with an allreduce_sum.
    matrix = O_LT @ O_LT.T
    matrix_side = matrix.shape[-1]  # * it can be ns or 2*ns, depending on mode

    shifted_matrix = jax.lax.add(
        matrix, diag_shift * jnp.eye(matrix_side, dtype=matrix.dtype)
    )
    # replicate

    if proj_reg is not None:
        shifted_matrix = jax.lax.add(
            shifted_matrix, jnp.full_like(shifted_matrix, proj_reg / matrix_side)
        )

    aus_vector = solver_fn(shifted_matrix, dv)

    # Some solvers return a tuple, some others do not.
    if isinstance(aus_vector, tuple):
        aus_vector, info = aus_vector
        if info is None:
            info = {}
    else:
        info = {}

    # (np, #ns) x (#ns) -> (np).
    updates = O_L.T @ aus_vector
    if momentum is not None:
        updates += momentum * old_updates
        old_updates = updates

    # If complex mode and we have complex parameters, we need
    # To repack the real coefficients in order to get complex updates
    if mode == "complex" and nkjax.tree_leaf_iscomplex(params_structure):
        num_p = updates.shape[-1] // 2
        updates = updates[:num_p] + 1j * updates[num_p:]

    if config.netket_experimental_sharding:
        out_shardings = (
            PositionalSharding(jax.devices()).replicate().reshape((1,) * updates.ndim)
        ).replicate()
        updates = jax.lax.with_sharding_constraint(updates, out_shardings)

    return updates, old_updates, info



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from netket.operator._discrete_operator import DiscreteOperator
from netket.operator._sum.base import SumOperator


class SumDiscreteOperator(SumOperator, DiscreteOperator):
    def __init__(self, *operators, coefficients=1.0, dtype=None):
        if not all(isinstance(op, DiscreteOperator) for op in operators):
            raise TypeError(
                "Arguments to SumDiscreteOperator must all be "
                "subtypes of DiscreteOperator. However the types are:\n\n"
                f"{list(type(op) for op in operators)}\n"
            )
        self._initialized = False
        super().__init__(
            operators, operators[0].hilbert, coefficients=coefficients, dtype=dtype
        )

    def _setup(self, force: bool = False):
        if not self._initialized:
            self._initialized = True

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        return sum(op.max_conn_size for op in self.operators)

    def get_conn_flattened(
        self,
        x: np.ndarray,
        sections: np.ndarray,
        pad: bool = False,
    ) -> tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError

    def to_jax_operator(self) -> "SumDiscreteJaxOperator":  # noqa: F821
        """
        Returns the standard (numba) version of this operator, which is an
        instance of {class}`nk.operator.Ising`.
        """
        from .discrete_jax_operator import SumDiscreteJaxOperator

        ops_numba = tuple(op.to_jax_operator() for op in self.operators)
        return SumDiscreteJaxOperator(
            *ops_numba, coefficients=self.coefficients, dtype=self.dtype
        )
